/*************************
 * Steven Stanek
 * March 18, 2001
 * yziformutil.c
 *************************/
 
#include "yziformutil.h"

#define	RANDOM_STR	"jaofmgq6toped89c.iff"
#define bool int
#define	true 1
#define	false 0
#define brev_32(x)  (x = (x<<24 & 0xFF000000) + (x<<8 & 0x00FF0000) + (x>>8 & 0x0000FF00) + (x>>24 & 0x000000FF))
#define brev_16(x)  (x = (x<<8 & 0xFF00) + (x>>8 & 0x00FF))

typedef struct
{
	int width, height;
	color_t **image_data;
} ImageData_t;

//Helpers for For colorArrtoImageFile
static void writeFile(ImageData_t *image, FILE *out);
static char* getNextID(char* last_ID);
static color_t *assembleOccuranceList(ImageData_t *image, int* colors);
static int	getNumOccurances(color_t *toFind, ImageData_t *image);
static bool nextOccuringPixel(color_t *toFind, ImageData_t *image, int *x, int *y);
static color_t *getNextColor(color_t *last_color, ImageData_t *image);
static int calcColorVal(color_t toCalc);


void colorArrtoImageFile(int width, int height, color_t** arr, FILE *out)
{
	ImageData_t image;
	image.width = width;
	image.height = height;
	image.image_data =arr;
	
	writeFile(&image, out);
}


static char* strdup2(char * todup)
{
	char * toreturn;
	
	assert(toreturn = malloc((strlen(todup) +1) * sizeof(char)));
	strcpy(toreturn, todup);
	return toreturn;
}



void colorArrtoBMP(color_t** image, FILE *out, int width, int height)
{
	unsigned int biWidth=width, biHeight=height;
	unsigned short bfType = 19778, bfReserved =0;
	unsigned int bfSize, bfOffset=54;
	unsigned short biPlanes=0, biBitCount = 24;
	unsigned int biSize = 40, biZero=0;
	int i, j;
	unsigned char r,g,b;

	bfSize = ((width*3) % 4 ==0) ? bfOffset+(3*width*height): bfOffset+(3*height*(width + 4 - (width*3) % 4));
	
	bfSize = (bfSize %4 == 0) ? (bfSize/4) *4:(bfSize/4) *4 +4;
	
	brev_16(bfType);
	brev_16(bfReserved);
	brev_32(bfSize);
	brev_32(bfOffset);
	brev_16(biPlanes);
	brev_16(biBitCount);
	brev_32(biSize);
	brev_32(biZero);
	brev_32(biWidth);
	brev_32(biHeight);
	
	fwrite(&bfType, 2, 1, out);
	fwrite(&bfSize, 4, 1, out);
	fwrite(&biZero, 4,1,out);
	fwrite(&bfOffset,4,1,out);
	fwrite(&biSize,4,1,out);
	fwrite(&biWidth,4,1,out);
	fwrite(&biHeight,4,1,out);
	fwrite(&biPlanes,2,1,out);
	fwrite(&biBitCount,2,1,out);
	fwrite(&biZero, 4,1,out);
	fwrite(&biZero, 4,1,out);
	fwrite(&biZero, 4,1,out);
	fwrite(&biZero, 4,1,out);
	fwrite(&biZero, 4,1,out);
	fwrite(&biZero, 4,1,out);
	
	for(i=height-1; i >= 0; i--)
	{
		for(j=0; j < width; j++)
		{
			r = (image[i][j].r) / 256;
			g = (image[i][j].g) / 256;
			b = (image[i][j].b) / 256;
			
			fwrite(&b,1,1,out);
			fwrite(&g,1,1,out);
			fwrite(&r,1,1,out);
		}
		
		for(j=0; width*3 % 4 != 0 &&  j < (4-(width*3)%4); j++)
		{
			r = 0;
			
			fwrite(&r,1,1,out);
		}
	}
}


void imageToBMP(yzImage dImage, FILE *out)
{
	int i;
	color_t** arr;
	
	arr=toColors(dImage);
	colorArrtoBMP(arr, out, getImageWidth(dImage), getImageHeight(dImage));
	
	for(i=0; i < getImageHeight(dImage); i++)
		free(arr[i]);
	free(arr);
}


color_t** readBMPFile(FILE *in, int *width, int *height)
{
	unsigned int intin;
	unsigned short shortin;
	unsigned char r, g, b;
	color_t **arr;
	int i, j;
	
	fread(&shortin,2,1,in);
	brev_16(shortin);
	
	if(shortin != 19778)
	{
		printf("Error: not a bitmap\n");
		return;
	}
	
	fread(&intin, 4,1,in);
	fread(&intin, 4,1,in);
	fread(&intin, 4,1,in);
	fread(&intin, 4,1,in);
	fread(&intin, 4,1,in);
	
	//Read width
	brev_32(intin);
	*width = intin;
	
	fread(&intin, 4,1,in);
	
	//Read height
	brev_32(intin);
	*height = intin;
	
	fread(&shortin,2,1,in);
	fread(&shortin,2,1,in);
	
	brev_16(shortin);
	
	if(shortin != 24)
	{
		printf("Error: Reader only supports 24 bit color\n");
		return;
	}
	
	fread(&intin, 4,1,in);
	
	brev_32(intin);
	
	if(intin != 0)
	{
		printf("This reader does not support compression\n");
		return;
	}
	
	fread(&intin, 4,1,in);
	fread(&intin, 4,1,in);
	fread(&intin, 4,1,in);
	fread(&intin, 4,1,in);
	fread(&intin, 4,1,in);
	
	assert(arr = malloc(*height * sizeof(color_t*)));
	
	for(i=0; i < *height; i++)
		assert(arr[i] = malloc(*width * sizeof(color_t)));
	
	for(i--; i >= 0; i--)
	{
		for(j=0; j < *width; j++)
		{
			fread(&b, 1, 1, in);
			fread(&g, 1, 1, in);
			fread(&r, 1, 1, in);
			
			arr[i][j].b = b * 256;
			arr[i][j].g = g * 256;
			arr[i][j].r = r * 256;
			arr[i][j].a=0;
		}
		
		//Read trailling zeros
		for(j=0; *width*3 % 4 != 0 &&  j < (4-(*width*3)%4); j++)
			fread(&r,1,1,in);
	}
	
	return arr;
}


void BMPfileToYZiFormFILE(FILE *bmpin, FILE *yziformout)
{
	int width, height, i;
	color_t **arr;
	
	arr=readBMPFile(bmpin, &width, &height);
	
	colorArrtoImageFile(width, height, arr, yziformout);
	
	for(i=0; i < height; i++)
		free(arr[i]);
	free(arr);
}


yzImage BMPtoyzImage(FILE *in)
{
	int width, height, i;
	color_t ** arr;
	yzImage dImage;
	FILE *out;
	
	arr=readBMPFile(in, &width, &height);
	
	out=fopen(RANDOM_STR, "w");
	colorArrtoImageFile(width, height, arr, out);
	fclose(out);
	
	out=fopen(RANDOM_STR, "r");
	dImage=openImageFile(out);
	fclose(out);
	
	for(i=0; i < height; i++)
		free(arr[i]);
	
	free(arr);
	
	remove(RANDOM_STR);
	
	return dImage;
}


/**********************************************************************************
 * 						HELPERS FOR colorArrtoImageFile
 **********************************************************************************/
static void writeFile(ImageData_t *image, FILE *out)
{
	int i, j, x, y, num_colors;
	char ***image_strs, *last_ID=0, *new_ID=0;
	color_t *dList;
	
	dList = assembleOccuranceList(image, &num_colors);
	assert(image_strs = malloc(image->height * sizeof(char**)));
	
	//Write out beginning of header
	fprintf(out, "<Begin YZiForm Image Version 0.0.0a>\n");
	fprintf(out, "<Definitions>\n");
	
	//Inits definitions array
	for(i=0; i < image->height; i++)
		assert(image_strs[i] = malloc(image->width * sizeof(char*)));
	
	//Read in defs by colors (most to least common) and outputs them
	for(i=0; i < num_colors; i++)
	{
		x=-1; y=0;
		
		new_ID = getNextID(last_ID);
		
		if(last_ID)
			free(last_ID);
			
		last_ID = strdup2(new_ID);
		
		//Adds leading and ending quotation to strings with length > 1
		if(strlen(last_ID) > 1)
		{
			assert(new_ID = malloc(strlen(last_ID) +3));
			strcpy(new_ID +1, last_ID);
			new_ID[strlen(last_ID) +1]= new_ID[0] = '\"';
			new_ID[strlen(last_ID) +2]= '\0';
		}
		
		//Writes a definition out
		fprintf(out, "%s = %d, %d, %d\n", new_ID, dList[i].r, dList[i].g, dList[i].b);
		
		//Adds this color to the map of the image
		while(nextOccuringPixel(&(dList[i]), image, &x, &y))
		{
			 image_strs[y][x] = strdup2(new_ID);
		}
		
		//Frees Strings associated with long quotes
		if(last_ID != new_ID)
			free(new_ID);
	}
	
	free(last_ID); //Frees last string
	
	//Outputs Image Header (end of defs)
	fprintf(out, "<Image %d,%d>\n", image->width, image->height);
	
	//Outputs the Image (row by row) & frees associated memory
	for(i=0; i < image->height; i++)
	{
		for(j=0; j < image->width; j++)
		{	
			if(image->image_data[i][j].a != TRANS_A)
				fprintf(out, "%s", image_strs[i][j]);
			else
				fprintf(out, "%s", TRANS_STR);
			
			free(image_strs[i][j]);
			
		}
		fprintf(out, "\n");
		free(image_strs[i]);
	}
	
	fprintf(out, "<End Image>");
	free(image_strs);
	free(dList);
}


//Pass NULL for beginning
static char* getNextID(char* last_ID)
{	
	int len, i;
	char *newStr;
	
	//If NULL return the starting string, the # 0
	if(!last_ID)
		return (char *) strdup2("0");
		
	len = strlen(last_ID);
		
	//Add another digit
	if(last_ID[len-1] == 'z' && last_ID[0] == 'z')
	{
		assert(newStr = malloc((len+2) *sizeof(char)));
			
		for(i=0; i < len +1; i++)
			newStr[i] = '0';
			
		newStr[len+1] = '\0';
		
		return newStr;
	}
	
	//If last digit = 'z' - roll back next non 'z' digit and replace previous
	// 'z's with '0's
	else
	{
		newStr = strdup2(last_ID);
		
		for(i=len-1; i >= 0; i--)
		{
			if(newStr[i] == 'z')
				newStr[i] = '0';
			
			else
			{	if(newStr[i] != ';')
					newStr[i] ++;
				else
					newStr[i] +=2;
				break;
			}
		}
				
		return newStr;
	}
}


/* Function: color_t *assembleOccuranceList(ImageData_t *image, int *colors)
 * This function assembles a list of all colors from most used to least used.
 * It does this by first determining the total number of colors in the image.
 * It then creates lists of color_ts and ints of that length. These lists are 
 * filled with colors and the corresponding number of times they appear in
 * the image. The lists are then selection sorted, most to least. Finally,
 * memory is freed, the num_colors is copied into the value at the passed in
 * colors pointer and the list of colors (in order) is returned.
 */
static color_t *assembleOccuranceList(ImageData_t *image, int *colors)
{
	int num_colors, i, j, *occurs, largest, temp, sum_of_colors=0;
	color_t *last=0, *list, *earlier=0, color_temp;
	
	for(i=0; 1; i++)
	{
		last= (i==0) ? getNextColor(0, image) : getNextColor(earlier, image);
		
		if(earlier != 0)
			free(earlier);
		
		if(last == 0)
			break;
			
		earlier = last;
	}
		
	num_colors = i;
	assert(list = malloc((num_colors+1) * sizeof(color_t)));
	assert(occurs = malloc(num_colors * sizeof(int)));
	
	for(i=0; i < num_colors; i++)
	{
		last = (i==0) ? getNextColor(0, image) : getNextColor(&(list[i-1]), image);
		list[i]= *last;
		free(last);
		occurs[i] = getNumOccurances(&(list[i]), image);
	}
	
	for(i=0; i < num_colors; i++)
	{
		largest = i;
		
		for(j=i; j < num_colors; j++)
		{
			if(occurs[j] > occurs[i])
				largest = j;
		}
		
		temp = occurs[i];
		color_temp = list[i];
		occurs[i] = occurs[largest];
		list[i] = list[largest];
		occurs[largest] = temp;
		list[largest] = color_temp;
	}
	
	//Error Checking
	for(i=0; i < num_colors; i++)
	{
		sum_of_colors += occurs[i];
	}
	
	if(sum_of_colors < image->width *image->height)
	{
		printf("You've been hit by the phantom bug. Please yell at Steven");
		exit(2);
	}
	
	*colors = num_colors;
	
	free(occurs);
	return list;
}


//Returns the number of times a color appears in the image
static int	getNumOccurances(color_t *toFind, ImageData_t *image)
{
	int i, x=-1, y=0;
	
	for(i=0; nextOccuringPixel(toFind, image, &x, &y); i++)
	{
		assert(1);
	}
	
	return i;
}


//Returns false if no more exist. Otherwise updates x & y to refer to the
//next pixel of that color.
static bool nextOccuringPixel(color_t *toFind, ImageData_t *image, int *x, int *y)
{
	(*x)++;

	for(; *y < image->height; (*y)++)
	{
		for(; *x < image->width; (*x)++)
		{
			if(image->image_data[*y][*x].r == toFind->r && image->image_data[*y][*x].g == toFind->g 
			    && image->image_data[*y][*x].b == toFind->b)
				return true;
		}
		*x=0;
	}
	return false;
}


/* Function: color_t *getNextColor(color_t *last_color, ImageData_t *image)
 * This function uses the calcColorVal function to find the color following
 * the given last_color. The color system ifself is an intricate ranking 
 * system. It automatically sets the lowest found color to max and the 
 * corresponding lowest_val variable to a number too high to determine if
 * no higher ranked colors exist. It uses a double loop to search for the
 * next highest color.
 */
static color_t *getNextColor(color_t *last_color, ImageData_t *image)
{
	int last_val, lowest_val, i, j, presCalc;
	color_t lowest_col, *toreturn;
		
	lowest_col.r = lowest_col.g = lowest_col.b = 65535;
	lowest_val = calcColorVal(lowest_col) +1;
	
	if(!last_color)
		last_val = -1;
	else
		last_val = calcColorVal(*last_color);
	
	
	for(i=0; i < image->height; i++)
	{
		for(j=0; j < image->width; j++)
		{
			presCalc = calcColorVal(image->image_data[i][j]);
			
			if(presCalc > last_val && presCalc < lowest_val)
			{
				lowest_val = presCalc;
				lowest_col = image->image_data[i][j];
			}
		}
	}
		
	if(lowest_val > calcColorVal(lowest_col))
		return 0;
	
	assert(toreturn = malloc(sizeof(color_t)));
	*toreturn = lowest_col;
	return toreturn;
}


//Creates a number which represents each color differently
static int calcColorVal(color_t toCalc)
{
	int toreturn;
	
	toreturn = toCalc.r/255;
	toreturn += toCalc.g/255 *256;
	toreturn += toCalc.b/255 *256 * 256;
	
	return toreturn;
}